/*
 * The Ladders with Broken Rungs
 *
 * main.c
 * Main module
 */

/* Copyright 2016, Marcin Peczarski. All Rights Reserved.
 *
 * This program is free software. You may use this software, and copy,
 * redistribute, and/or modify it for non profit works.
 * Please do not remove this copyright notice.
 */

#include "main.h"

/*
 * Global data initialization and deinitialization
 */

#ifdef _USE_POF2_TABLE_
/* One element set - Power OF 2 table */
SET pof2_tbl[N];
#endif

static void globals_init(void) {
  #ifdef _USE_POF2_TABLE_
  int i;
  SET p;

  for (i = 0, p = 1; i < N; ++i, p <<= 1)
    pof2_tbl[i] = p;
  #endif
}

static void main_init(void) {
  globals_init();
  /* cple_module_init(); */
}

static void main_clean(void) {
  /* cple_module_clean(); */
}

/*
 * Poset routines
 */

static void poset_add_pair(POSET *g, int v1, int v2) {
  int v;
  SET v_set, n_set;

  /* Add pair (v1,v2) to r[0]. All elements till now greater than v2 are now
     greater than v1. If v was lower than v1, then now v is lower than all
     new elements greater than v1. */
  n_set = g->r[0][v1] |= pof2(v2) | g->r[0][v2];
  v_set = g->r[1][v1];
  while (v_set) {
    v = min_element(v_set);
    v_set ^= pof2(v);
    g->r[0][v] |= n_set;
  }

  /* Add pair (v2,v1) to r[1]. All elements till now lower than v1 are now
     lower than v2. If v was greater than v2, then now v is greater than all
     new elements lower than v2. */
  n_set = g->r[1][v2] |= pof2(v1) | g->r[1][v1];
  v_set = g->r[0][v2];
  while (v_set) {
    v = min_element(v_set);
    v_set ^= pof2(v);
    g->r[1][v] |= n_set;
  }
}

static void compose_ladder(POSET *poset, SET mask) {
  int i;

  poset_empty(poset);
  poset->n = N;
  for (i = 0; i <= N - 3; ++i) /* two chains */
    poset_add_pair(poset, i, i + 2);
  for (i = 0; i <= N - 4; ++i) /* rungs */
    if (mask & pof2(i))
      poset_add_pair(poset, i, i + 3);
}

static void print_ladder(SET mask) {
  int i;

  printf("L_{%d", N);
  for (i = 0; i <= N - 4; ++i)
    if (!(mask & pof2(i)))
      printf(",%d", i);
  printf("}");
}

/*
 * Atkinson-Chang algorithm for counting linear extensions
 */

/*
x_1, x_2, ..., x_p are even elements
y_1, y_2, ..., y_q are odd elements
p = (N + 1)/2
q = N/2
a_i is a[i - 1]
b_i is b[i - 1], b_0 = 0
*/
static CPC ach(POSET const *poset) {
  static const int p = (N + 1)/2;
  static const int q = N/2;

  CPC t, u[q + 1];
  int a[p], b[p], i, j;
  SET A, B;

  for (i = 0; i < p; ++i) {
    A = poset->r[0][i << 1] & ODD_SET;
    B = poset->r[1][i << 1] & ODD_SET;
    a[i] = A ? (min_element(A) >> 1) + 1 : q + 1;
    b[i] = B ? (max_element(B) >> 1) + 1 : 0;
  }
  /* The value of u[0] is initialized by 1.
     The first iteration for i = 0 is unrolled (we have b[-1] + 1 == 1)
       for (j = b[-1] + 1; j < a[0]; ++j) */
  for (j = 0; j < a[0]; ++j)
    u[j] = 1;
  for (j = a[0]; j <= q; ++j)
    u[j] = 0;
  for (i = 1; i < p; ++i)
    for (j = b[i - 1] + 1; j < a[i]; ++j)
      u[j] += u[j - 1];
  t = 0;
  for (j = b[p - 1]; j < a[p - 1]; ++j)
    t += u[j];

  return t;
}

/*
 * Main experiment routine
 */

static CPC balance_constant_compute(SET mask, CPC t0) {
  POSET p0, p1;
  CPC   tmax, t1;
  int   i, j;

  compose_ladder(&p0, mask);
  tmax = 0;
  for (i = 0; i < N; i += 2) {
    for (j = 1; j < N; j += 2) {
      if (!(p0.r[0][i] & pof2(j))) {
        p1 = p0;
        poset_add_pair(&p1, i, j);
        t1 = ach(&p1);
        if (2 * t1 > t0)
          t1 = t0 - t1;
        if (t1 > tmax)
          tmax = t1;
      }
    }
  }
  return tmax;
}

#define MAX_RESULT_COUNT  2

/* Search for ladders with balance constant less than numerator/denominator. */
static int balance_constant_check(CPC numerator, CPC denominator, int type) {
  POSET poset;
  CPC t0, t1, min_numerator[MAX_RESULT_COUNT], min_denominator[MAX_RESULT_COUNT];
  SET m, mask, M, min_mask[MAX_RESULT_COUNT];
  long double b, min_b;
  long generated, analysed;
  int i, result_count;

  printf("searching for balance constant < " CPC_PRINT_FORMAT "/" CPC_PRINT_FORMAT "   type %d\n",
         numerator, denominator, type);

  if (type == 0) {
    /* all ladders with broken rungs */
    M = 1L << (N - 3);
  }
  else if (type == 1) {
    /* L_{1,5,...,n-9,n-5} */
    #if N >= 15
    M = 1UL << (N - 15);
    #else
    return 1;
    #endif
  }
  else if (type == 2) {
    /* L_{1,5,8,...,n-12,n-9,n-5} */
    #if N >= 21
    M = 1UL << (N - 21);
    #else
    return 1;
    #endif
  }
  else {
    return 1;
  }

  generated = analysed = 0;
  result_count = 0;
  min_b = (long double)numerator/(long double)denominator;
  for (m = 0; m < M; ++m) { /* mask is rung mask */
    ++generated;
    mask = m;
    if (type == 1) {
      #if N >= 15
      mask <<= 6;
      mask |= 0x1DUL | 0x2EUL << (N - 9);
      #endif
    }
    else if (type == 2) {
      #if N >= 21
      mask <<= 9;
      mask |=  0xDDUL | 0x176UL << (N - 12);
      #endif
    }
    compose_ladder(&poset, mask);
    t0 = ach(&poset);
    poset_add_pair(&poset, 0, 1);
    t1 = ach(&poset);
    if (2 * t1 > t0) {
      t1 = t0 - t1;
    }
    b = (long double)t1/(long double)t0;
    if (b < min_b) {
      ++analysed;
      t1 = balance_constant_compute(mask, t0);
      b = (long double)t1/(long double)t0;
      if (b < min_b) {
        min_b = b;
        min_numerator[0] = t1;
        min_denominator[0] = t0;
        min_mask[0] = mask;
        result_count = 1;
      }
      else if (b == min_b) {
        if (result_count < MAX_RESULT_COUNT) {
          min_numerator[result_count] = t1;
          min_denominator[result_count] = t0;
          min_mask[result_count] = mask;
          ++result_count;
        }
        else {
          fprintf(stderr, "ERROR: too small value of MAX_RESULT_COUNT.\n");
          return 1;
        }
      }
    }
  }

  for (i = 0; i < result_count; ++i) {
    printf("b(");
    print_ladder(min_mask[i]);
    printf(") = " CPC_PRINT_FORMAT "/" CPC_PRINT_FORMAT " = %.15Lf\n",
           min_numerator[i], min_denominator[i],
           (long double)min_numerator[i]/(long double)min_denominator[i]);
  }
  printf("generated posets %12ld\n"
         "analysed posets  %12ld\n"
         "output posets    %12d\n",
         generated, analysed, result_count);

  return 0;
}

int main(int argc, char *argv[]) {
  long timer;
  int ret;

  if (argc != 4) {
    printf("usage: %s numerator  denominator search_range\n", argv[0]);
    return 3;
  }

  timer = time(NULL);

  main_init();
  ret = balance_constant_check(atol(argv[1]), atol(argv[2]), atoi(argv[3]));
  main_clean();

  printf("computation time %12ld\n"
         "\n",
         time(NULL) - timer);

  return ret;
}
