/*
 * The Worst Balanced Posets
 *
 * wbp.c
 * Hash table, experiment routines, etc.
 */

/* Copyright 2002--2015, Marcin Peczarski. All Rights Reserved.
 *
 * This program is free software. You may use this software, and copy,
 * redistribute, and/or modify it for non profit works.
 * Please do not remove this copyright notice.
 */

#include "main.h"

/* Tables speed up counting hash functions */
static INDEX h1_tbl[N][N];

#ifdef _USE_POF2_TABLE_
/* One element set - Power OF 2 table */
SET pof2_tbl[N];
#endif

#ifdef _USE_MIN_ELEMENT_TABLE_
/* Lowest element in set */
MIN_EL_TBL_EL min_element_tbl[MET_SIZE + 384];
#endif

#ifdef _USE_CARDINALITY_TABLE_
/* Set cardinality - counting ones */
CARD_TBL_EL cardinality_tbl[CAT_SIZE + 384];
#endif

/* Hash tables and stacks for posets storage */
#define EMPTY_POS  -1
static INDEX  new_posets_count;
static INDEX  hash_tbl[HASH_TBL_SIZE];
static POSET  posets_stack[MAX_STACK];

void wbp_module_init(void) {
  INDEX w1[(N + 1) * (N + 1) / 4];
  int   i, m1, m2;

  #ifdef _USE_POF2_TABLE_
  SET p;

  for (i = 0, p = 1; i < N; ++i, p <<= 1)
    pof2_tbl[i] = p;
  #endif

  #ifdef _USE_MIN_ELEMENT_TABLE_
  min_element_tbl[0] = 0;
  min_element_tbl[1] = 0;
  i = 1;
  while (i < MET_SIZE) {
    i++;
    /* i is even */
    min_element_tbl[i] = min_element_tbl[i >> 1] + 1;
    i++;
    /* i is odd */
    min_element_tbl[i] = 0;
  }
  #endif

  #ifdef _USE_CARDINALITY_TABLE_
  cardinality_tbl[0] = 0;
  cardinality_tbl[1] = 1;
  i = 1;
  while (i < CAT_SIZE) {
    i++;
    /* i is even */
    cardinality_tbl[i] = cardinality_tbl[i >> 1];
    i++;
    /* i is odd */
    cardinality_tbl[i] = cardinality_tbl[i >> 1] + 1;
  }
  #endif

  w1[0] = 1;
  for (i = 1; i < (N + 1) * (N + 1) / 4; ++i)
    w1[i] = (w1[i - 1] * N) % HASH_TBL_SIZE;
  for (m2 = 0; m2 < N; ++m2)
    for (m1 = 0; m1 <= m2; ++m1)
      if (m1 + m2 < N)
        h1_tbl[m1][m2] = h1_tbl[m2][m1] = w1[m1 * (N - m1) + m2];
      else /* These values are never used. */
        h1_tbl[m1][m2] = h1_tbl[m2][m1] = 0;
}

void wbp_module_clean(void) {
}

static void empty_storage() {
  INDEX i;

  new_posets_count = 0;
  for (i = 0; i < HASH_TBL_SIZE; ++i)
    hash_tbl[i] = EMPTY_POS;
}

static INDEX search(POSET *g) {
  INDEX h;
  int   i;

  h = 0;
  for (i = 0; i < N; ++i)
    h += h1_tbl[cardinality(g->r[1][i])][cardinality(g->r[0][i])];
  h %= HASH_TBL_SIZE;
  g->hash = h;
  for (h = hash_tbl[h]; h != EMPTY_POS; h = posets_stack[h].hash) {
    if (posets_stack[h].n == g->n) {
      if (congruent(posets_stack + h, g)) {
        return h;
      }
    }
  }
  return EMPTY_POS;
}

static void insert(POSET const *g) {
  if (new_posets_count >= MAX_STACK)
    fatal_error("insert, posets stack overflow");

  posets_stack[new_posets_count] = *g;
  posets_stack[new_posets_count].hash = hash_tbl[g->hash];
  hash_tbl[g->hash] = new_posets_count;
  new_posets_count++;
}

static void poset_add_pair(POSET *g, int v1, int v2) {
  int v;
  SET v_set, n_set;

  /* Add pair (v1,v2) to r[0]. All elements till now greater than v2 are now
     greater than v1. If v was lower than v1, then now v is lower than all
     new elements greater than v1. */
  n_set = g->r[0][v1] |= pof2(v2) | g->r[0][v2];
  v_set = g->r[1][v1];
  while (v_set) {
    v = min_element(v_set);
    v_set ^= pof2(v);
    g->r[0][v] |= n_set;
  }

  /* Add pair (v2,v1) to r[1]. All elements till now lower than v1 are now
     lower than v2. If v was greater than v2, then now v is greater than all
     new elements lower than v2. */
  n_set = g->r[1][v2] |= pof2(v1) | g->r[1][v1];
  v_set = g->r[0][v2];
  while (v_set) {
    v = min_element(v_set);
    v_set ^= pof2(v);
    g->r[1][v] |= n_set;
  }
}

/*
 * Compare two fractions. This works well if CPC is 32-bit int and long is 64-bit.
 */

typedef enum {
  LESS = -1, EQUAL = 0, GREATER = 1
} cmp_result_t;

typedef struct {
  CPC numerator, denominator;
} fraction_t;

static cmp_result_t cmp(fraction_t left, fraction_t right) {
  long left_f, right_f;

  left_f = (long)left.numerator * (long)right.denominator;
  right_f = (long)right.numerator * (long)left.denominator;
  if (left_f < right_f)
    return LESS;
  else if (left_f > right_f)
    return GREATER;
  else
    return EQUAL;
}

/*
 * Main experiment routines
 */

static BOOL wbp_check_not_linear_sum(POSET *p) {
  int j;
  SET A, B, V;

  for (B = 1; B <= FULL_SET / 2; B++) {
    V = B;
    A = FULL_SET;
    while (V) {
      j = min_element(V);
      A &= p->r[0][j] | p->r[1][j];
      V ^= pof2(j);
      if ((A | B) != FULL_SET)
        goto next_B_set; /* There is no hope that (A ^ B) == FULL_SET. */
    }
    if ((A ^ B) == FULL_SET)
      return FALSE; /* a linear sum */
next_B_set:
    ;
  }
  return TRUE; /* not a linear sum */
}

static void  wbp_search_and_insert(POSET *p) {
  /* The condition is lazy evaluated. Calling search before
     wbp_check_not_linear_sum shortens computation time. */
  if (search(p) == EMPTY_POS && wbp_check_not_linear_sum(p))
    insert(p);
}

typedef struct {
  fraction_t b;
  int        count, j[N], k[N];
} wbp_info_t;

static BOOL wbp_test(POSET const *p, wbp_info_t *wbp_info) {
  CPC t[N][N];
  int j, k;

  if (N > 3 && p->n < N) {
    /* This poset was considered in test for N - 1 elements. */
    return FALSE;
  }

  wbp_info->b.numerator = 0;
  wbp_info->b.denominator = lec(p, t);
  wbp_info->count = 0;
  for (j = 0; j < N - 1; j++) {
    for (k = j + 1; k < N; k++) {
      if (t[j][k] == t[k][j]) {
        /* We are not interested in posets with b(P) = 1/2. */
        return FALSE;
      }
      else if (t[j][k] > t[k][j]) {
        if (t[k][j] > wbp_info->b.numerator) {
          wbp_info->b.numerator = t[k][j];
          wbp_info->j[0] = k;
          wbp_info->k[0] = j;
          wbp_info->count = 1;
        }
        else if (t[k][j] == wbp_info->b.numerator && wbp_info->count < N) {
          wbp_info->j[wbp_info->count] = k;
          wbp_info->k[wbp_info->count] = j;
          wbp_info->count++;
        }
      }
      else {
        if (t[j][k] > wbp_info->b.numerator) {
          wbp_info->b.numerator = t[j][k];
          wbp_info->j[0] = j;
          wbp_info->k[0] = k;
          wbp_info->count = 1;
        }
        else if (t[j][k] == wbp_info->b.numerator && wbp_info->count < N) {
          wbp_info->j[wbp_info->count] = j;
          wbp_info->k[wbp_info->count] = k;
          wbp_info->count++;
        }
      }
    }
  }

  return TRUE;
}

#define WBP_MAX_COUNT 8

int wbp_iterative_search(int check_type) {
  POSET        *p0, p1, wbp[WBP_MAX_COUNT];
  wbp_info_t   wbp_info[WBP_MAX_COUNT], wbp_tmp_info;
  fraction_t   b;
  cmp_result_t res;
  long         timer;
  INDEX        l, from, to;
  int          i, j, k, m, n, step, wbp_count;
  SET          A;

  timer = time(NULL);
  empty_storage();
  step = 0;
  to = 0;

  switch (check_type) {
    case 0: /* not a linear sum */
      poset_empty(&p1);
      /* p1.n = 0; This is set by the poset_empty function. */
      wbp_search_and_insert(&p1);
      break;
    case 1: /* ladders with broken rungs */
      for (m = 0; m < (1 << (N - 3)); ++m) { /* m is rung mask */
        poset_empty(&p1);
        p1.n = N;
        for (i = 0; i <= N - 3; ++i) /* two chains */
          poset_add_pair(&p1, i, i + 2);
        for (i = 0; i <= N - 4; ++i) /* rungs */
          if (m & (1 << i))
            poset_add_pair(&p1, i, i + 3);
        if (search(&p1) == EMPTY_POS) /* Ladder is never a linear sum. */
          insert(&p1);
      }
      break;
    case 2: /* not a linear sum, width 2 */
      for (m = 1; m <= N/2; ++m) {
        poset_empty(&p1);
        p1.n = N;
        for (i = 1; i < m; ++i)
          poset_add_pair(&p1, i - 1, i);
        for (i = m + 1; i < N; ++i)
          poset_add_pair(&p1, i - 1, i);
        wbp_search_and_insert(&p1);
      }
      break;
    case 3: /* not a linear sum, width at most 3 */
      for (j = 1; j < N; ++j) {
        for (k = j; k < N; ++k) {
          poset_empty(&p1);
          p1.n = N;
          for (i = 1; i < N; ++i)
            if (i != j && i != k)
              poset_add_pair(&p1, i - 1, i);
          wbp_search_and_insert(&p1);
        }
      }
      break;
    default:
      fatal_error("wbp_iterative_search, not implemented check_type %d", check_type);
  }

  printf("Generating %d-element posets\n"
         "step %d  input posets %ld  output posets %ld\n",
         N, step, (long)to, (long)new_posets_count);
  if (check_type != 1) {
    while (to < new_posets_count && new_posets_count <= MAX_STACK) {
      step++;
      from = to;
      to = new_posets_count;
      for (l = from; l < to; ++l) {
        p0 = posets_stack + l;
        n = p0->n;
        if (n < N - 1) {
          p1 = *p0;
          p1.n = n + 2;
          poset_add_pair(&p1, n, n + 1);
          wbp_search_and_insert(&p1);
        }
        if (n < N) {
          for (j = 0; j < n; ++j) {
            p1 = *p0;
            p1.n = n + 1;
            poset_add_pair(&p1, j, n);
            wbp_search_and_insert(&p1);
            p1 = *p0;
            p1.n = n + 1;
            poset_add_pair(&p1, n, j);
            wbp_search_and_insert(&p1);
          }
        }
        for (j = 0; j < n - 1; ++j) {
          /* Set A contains elements not comparable with element j and such that k > j. */
          A = ~(p0->r[0][j] | p0->r[1][j]) & (FULL_SET >> (N - n)) & (FULL_SET << (j + 1));
          while (A) {
            k = min_element(A);
            A ^= pof2(k);
            p1 = *p0;
            poset_add_pair(&p1, j, k);
            wbp_search_and_insert(&p1);
            p1 = *p0;
            poset_add_pair(&p1, k, j);
            wbp_search_and_insert(&p1);
          }
        }
      }
      printf("step %d  input posets %ld  output posets %ld\n",
             step, (long)to, (long)new_posets_count);
    }
  }
  printf("computation time  %9ld\n\n", time(NULL) - timer);

  b.numerator = 1;
  b.denominator = 1;
  wbp_count = 0;
  for (l = 0; l < new_posets_count; ++l) {
    if (wbp_test(posets_stack + l, &wbp_tmp_info)) {
      res = cmp(wbp_tmp_info.b, b);
      if (res == EQUAL) {
        if (wbp_count < WBP_MAX_COUNT) {
          wbp[wbp_count] = *(posets_stack + l);
          wbp_info[wbp_count] = wbp_tmp_info;
          wbp_count++;
        }
      }
      else if (res == LESS) {
        wbp_count = 1;
        b = wbp_tmp_info.b;
        wbp[0] = *(posets_stack + l);
        wbp_info[0] = wbp_tmp_info;
      }
    }
  }
  printf("The Worst Balanced Poset\n");
  for (j = 0; j < wbp_count; ++j) {
    poset_print(wbp + j);
    printf("b(P) = " CPC_PRINT_FORMAT "/" CPC_PRINT_FORMAT " = %.15Lf for pairs",
           wbp_info[j].b.numerator, wbp_info[j].b.denominator,
           (long double)wbp_info[j].b.numerator / (long double)wbp_info[j].b.denominator);
    for (i = 0; i < wbp_info[j].count; ++i)
      printf(" (%d,%d)", wbp_info[j].j[i], wbp_info[j].k[i]);
  }
  printf("\n"
         "analysed posets   %9ld\n"
         "computation time  %9ld\n"
         "\n",
         (long)new_posets_count,
         time(NULL) - timer);

  return 0;
}
